'use strict';

var windGlCore = require('wind-gl-core');
var maptalks = require('maptalks');
var windCore = require('wind-core');
var rewind = require('@mapbox/geojson-rewind');
var visEngine = require('@sakitam-gis/vis-engine');

function _interopNamespaceDefault(e) {
  var n = Object.create(null);
  if (e) {
    Object.keys(e).forEach(function (k) {
      if (k !== 'default') {
        var d = Object.getOwnPropertyDescriptor(e, k);
        Object.defineProperty(n, k, d.get ? d : {
          enumerable: true,
          get: function () { return e[k]; }
        });
      }
    });
  }
  n.default = e;
  return Object.freeze(n);
}

var maptalks__namespace = /*#__PURE__*/_interopNamespaceDefault(maptalks);

function containsCoordinate(extent, coordinate) {
  const x = coordinate[0];
  const y = coordinate[1];
  return extent[0] <= x && x <= extent[2] && extent[1] <= y && y <= extent[3];
}
function transform(input, opt_output, opt_dimension) {
  const length = input.length;
  const dimension = opt_dimension > 1 ? opt_dimension : 2;
  let output = opt_output;
  if (output === void 0) {
    if (dimension > 2) {
      output = input.slice();
    } else {
      output = new Array(length);
    }
  }
  for (let i = 0; i < length; i += dimension) {
    output[i] = 180 * input[i] / 20037508342789244e-9;
    output[i + 1] = 360 * Math.atan(Math.exp(input[i + 1] / 6378137)) / Math.PI - 90;
  }
  return output;
}
function createOrUpdate(minX, minY, maxX, maxY, opt_extent) {
  if (opt_extent) {
    opt_extent[0] = minX;
    opt_extent[1] = minY;
    opt_extent[2] = maxX;
    opt_extent[3] = maxY;
    return opt_extent;
  } else {
    return [minX, minY, maxX, maxY];
  }
}
function boundingExtentXYs(xs, ys, opt_extent) {
  const minX = Math.min.apply(null, xs);
  const minY = Math.min.apply(null, ys);
  const maxX = Math.max.apply(null, xs);
  const maxY = Math.max.apply(null, ys);
  return createOrUpdate(minX, minY, maxX, maxY, opt_extent);
}
function applyTransform(extent, transformFn, opt_extent, opt_stops) {
  let coordinates = [];
  if (opt_stops > 1) {
    const width = extent[2] - extent[0];
    const height = extent[3] - extent[1];
    for (let i = 0; i < opt_stops; ++i) {
      coordinates.push(
        extent[0] + width * i / opt_stops,
        extent[1],
        extent[2],
        extent[1] + height * i / opt_stops,
        extent[2] - width * i / opt_stops,
        extent[3],
        extent[0],
        extent[3] - height * i / opt_stops
      );
    }
  } else {
    coordinates = [extent[0], extent[1], extent[2], extent[1], extent[2], extent[3], extent[0], extent[3]];
  }
  transformFn(coordinates, coordinates, 2);
  const xs = [];
  const ys = [];
  for (let i = 0, l = coordinates.length; i < l; i += 2) {
    xs.push(coordinates[i]);
    ys.push(coordinates[i + 1]);
  }
  return boundingExtentXYs(xs, ys, opt_extent);
}
function transformExtent(extent, opt_stops) {
  return applyTransform(extent, transform, void 0, opt_stops);
}

const _options = {
  renderer: "canvas",
  doubleBuffer: false,
  animation: false,
  windOptions: {}
};
class WindLayerRenderer extends maptalks.renderer.CanvasLayerRenderer {
  checkResources() {
    return [];
  }
  getDrawParams() {
    return [];
  }
  hitDetect() {
    return false;
  }
  draw() {
    this.prepareCanvas();
    this.prepareDrawContext();
    this.drawWind();
  }
  _redraw() {
    this.prepareRender();
    this.draw();
  }
  drawWind() {
    const map = this.getMap();
    if (this.context) {
      const layer = this.layer;
      const opt = layer.getWindOptions();
      if (!this.wind && map) {
        const data = layer.getData();
        this.wind = new windCore.WindCore(this.context, opt, data);
        this.wind.project = this.project.bind(this);
        this.wind.unproject = this.unproject.bind(this);
        this.wind.intersectsCoordinate = this.intersectsCoordinate.bind(this);
        this.wind.postrender = () => {
          this.setCanvasUpdated();
        };
        this.wind.prerender();
      }
      if (this.wind) {
        this.wind.prerender();
        this.wind.render();
      }
    }
    this.completeRender();
  }
  project(coordinate) {
    const map = this.getMap();
    const pixel = map.coordinateToContainerPoint(new maptalks.Coordinate(coordinate[0], coordinate[1]));
    return [pixel.x, pixel.y];
  }
  unproject(pixel) {
    const map = this.getMap();
    const coordinates = map.containerPointToCoordinate(new maptalks.Point(pixel[0], pixel[1]));
    return coordinates.toArray();
  }
  intersectsCoordinate(coordinate) {
    if (!this.layer)
      return false;
    const proj = this.layer?.getProjection();
    if (proj && proj.code !== "EPSG:3857") {
      return true;
    }
    const map = this.getMap();
    const projExtent = map.getProjExtent();
    const extent = [projExtent.xmin, projExtent.ymin, projExtent.xmax, projExtent.ymax];
    const mapExtent = transformExtent(extent, 0);
    return containsCoordinate(mapExtent, [coordinate[0], coordinate[1]]);
  }
  drawOnInteracting() {
  }
  onZoomStart(...args) {
    if (this.wind) {
      this.wind.stop();
    }
    super.onZoomStart.apply(this, args);
  }
  onZoomEnd(...args) {
    if (this.wind) {
      this.wind.start();
    }
    super.onZoomEnd.apply(this, args);
  }
  onDragRotateStart(...args) {
    if (this.wind) {
      this.wind.stop();
    }
    super.onDragRotateStart.apply(this, args);
  }
  onDragRotateEnd(...args) {
    if (this.wind) {
      this.wind.start();
    }
    super.onDragRotateEnd.apply(this, args);
  }
  onMoveStart(...args) {
    if (this.wind) {
      this.wind.stop();
    }
    super.onMoveStart.apply(this, args);
  }
  onMoveEnd(...args) {
    if (this.wind) {
      this.wind.start();
    }
    super.onMoveEnd.apply(this, args);
  }
  // onResize() {}
  remove() {
    if (this.wind) {
      this.wind.stop();
    }
    delete this._drawContext;
    super.remove();
  }
  getMap() {
    return super.getMap();
  }
  prepareCanvas() {
    return super.prepareCanvas();
  }
  prepareDrawContext() {
    super.prepareDrawContext();
  }
  prepareRender() {
    return super.prepareRender();
  }
  completeRender() {
    return super.completeRender();
  }
}
class MaptalksWind extends maptalks.CanvasLayer {
  constructor(id, data, options = {}) {
    super(id, windCore.assign({}, _options, options));
    this._map = null;
    this.pickWindOptions();
    if (data) {
      this.setData(data, options.fieldOptions);
    }
  }
  // onAdd() {}
  // onRemove() {}
  pickWindOptions() {
    Object.keys(windCore.defaultOptions).forEach((key) => {
      if (this.options && key in this.options) {
        if (this.options.windOptions === void 0) {
          this.options.windOptions = {};
        }
        this.options.windOptions[key] = this.options[key];
      }
    });
  }
  /**
   * get wind layer data
   */
  getData() {
    return this.field;
  }
  /**
   * set layer data
   * @param data
   * @param options
   * @returns {WindLayer}
   */
  setData(data, options = {}) {
    if (data && data.checkFields && data.checkFields()) {
      this.field = data;
    } else if (windCore.isArray(data)) {
      this.field = windCore.formatData(data, options);
    } else {
      console.error("Illegal data");
    }
    const r = this._getRenderer();
    if (r && r.wind && this.field) {
      r.wind.updateData(this.field);
    }
    return this;
  }
  setWindOptions(options) {
    const beforeOptions = this.options.windOptions || {};
    this.options = windCore.assign(this.options, {
      windOptions: windCore.assign(beforeOptions, options || {})
    });
    const r = this._getRenderer();
    if (r && r.wind) {
      const windOptions = this.options.windOptions;
      r.wind.setOptions(windOptions);
    }
  }
  getWindOptions() {
    return this.options.windOptions || {};
  }
  draw() {
    if (this._getRenderer()) {
      this._getRenderer()._redraw();
    }
    return this;
  }
  prepareToDraw() {
    return [];
  }
  drawOnInteracting() {
    this.draw();
  }
  _getRenderer() {
    return super._getRenderer();
  }
}
MaptalksWind.registerRenderer("canvas", WindLayerRenderer);

visEngine.highPrecision(true);
function getGLRes(map) {
  return map.getGLRes ? map.getGLRes() : map.getGLZoom();
}
function inRange(value, start, end) {
  return value >= start && value < end;
}
function coordinateToPoint(map, coordinate, res, out) {
  if (map.coordToPointAtRes) {
    return map.coordToPointAtRes(coordinate, res, out);
  }
  return map.coordinateToPoint(coordinate, res, out);
}
const TILE_POINT = new maptalks__namespace.Point(0, 0);
class VeRenderer extends maptalks__namespace.renderer.CanvasLayerRenderer {
  #renderTarget;
  getPrepareParams() {
    return [this.scene, this.camera];
  }
  getDrawParams() {
    return [this.scene, this.camera];
  }
  _drawLayer(...args) {
    super._drawLayer.apply(this, args);
  }
  hitDetect() {
    return false;
  }
  createCanvas() {
    super.createCanvas();
    this.createContext();
  }
  createContext() {
    if (this.canvas.gl && this.canvas.gl.wrap) {
      this.gl = this.canvas.gl.wrap();
    } else {
      const layer = this.layer;
      const attributes = layer.options.glOptions || {
        alpha: true,
        depth: true,
        antialias: true,
        stencil: true
      };
      attributes.preserveDrawingBuffer = true;
      this.gl = this.gl || visEngine.utils.getContext(this.canvas, attributes, layer.options?.requestWebGl2);
    }
    this.#initRenderer();
    this.layer.onCanvasCreate(this.context, this.scene, this.camera);
  }
  #initRenderer() {
    const renderer = new visEngine.Renderer(this.gl, {
      autoClear: false,
      extensions: [
        "OES_texture_float",
        "OES_texture_float_linear",
        "WEBGL_color_buffer_float",
        "EXT_color_buffer_float"
      ]
    });
    renderer.setSize(this.canvas.width, this.canvas.height);
    this.context = renderer;
    this.context.state.setClearColor({
      r: 0,
      g: 0,
      b: 0,
      a: 0
    });
    this.scene = new visEngine.Scene();
    const map = this.layer.getMap();
    const fov = map.getFov() * Math.PI / 180;
    this.camera = new visEngine.PerspectiveCamera(fov, map.width / map.height, map.cameraNear, map.cameraFar);
    this.planeCamera = new visEngine.OrthographicCamera(0, 1, 1, 0, 0, 1);
    this.camera.matrixAutoUpdate = false;
    this.#syncCamera();
  }
  onCanvasCreate() {
    super.onCanvasCreate();
  }
  resizeCanvas(canvasSize) {
    if (!this.canvas) {
      return;
    }
    let size;
    const map = this.getMap();
    if (!canvasSize) {
      size = map.getSize();
    } else {
      size = canvasSize;
    }
    const r = map.getDevicePixelRatio ? map.getDevicePixelRatio() : maptalks__namespace.Browser.retina ? 2 : 1;
    const canvas = this.canvas;
    const { width, height, cssWidth, cssHeight } = maptalks__namespace.Util.calCanvasSize(size, r);
    if (this.layer._canvas && (canvas.style.width !== cssWidth || canvas.style.height !== cssHeight)) {
      canvas.style.width = cssWidth;
      canvas.style.height = cssHeight;
    }
    if (canvas.width === width && canvas.height === height) {
      return this;
    }
    canvas.width = width;
    canvas.height = height;
    this.context.setSize(canvas.width, canvas.height);
  }
  clearCanvas() {
    if (!this.canvas) {
      return;
    }
    this.context.clear();
  }
  prepareCanvas() {
    if (!this.canvas) {
      this.createCanvas();
    } else {
      this.clearCanvas();
    }
    this.layer.fire("renderstart", { context: this.context });
    return null;
  }
  renderScene(context) {
    this.scene.worldMatrixNeedsUpdate = true;
    this.#syncCamera();
    if (context && context.renderTarget) {
      const { width, height } = context.renderTarget.fbo;
      if (!this.#renderTarget) {
        this.#renderTarget = new visEngine.RenderTarget(this.context, {
          width,
          height,
          depth: false
        });
      } else {
        this.#renderTarget.resize(width, height);
      }
      this.layer.layer?.prerender(
        {
          camera: this.camera,
          planeCamera: this.planeCamera
        }
        // this.#renderTarget,
      );
      this.#renderTarget.swapHandle(context.renderTarget.getFramebuffer(context.renderTarget.fbo));
      this.layer.layer?.render(
        {
          camera: this.camera,
          planeCamera: this.planeCamera
        },
        this.#renderTarget
      );
      this.#renderTarget.restoreHandle();
    } else {
      this.layer.layer?.prerender({
        camera: this.camera,
        planeCamera: this.planeCamera
      });
      this.layer.layer?.render({
        camera: this.camera,
        planeCamera: this.planeCamera
      });
    }
    this.completeRender();
  }
  remove() {
    delete this._drawContext;
    if (this.#renderTarget) {
      this.#renderTarget.destroy();
      this.#renderTarget = void 0;
    }
    super.remove();
  }
  #syncCamera() {
    const map = this.getMap();
    const camera = this.camera;
    camera.localMatrix.fromArray(map.cameraWorldMatrix);
    camera.projectionMatrix.fromArray(map.projMatrix);
    camera.worldMatrixNeedsUpdate = true;
    camera.updateMatrixWorld();
  }
  isTileCachedOrLoading() {
    return false;
  }
}
const options = {
  renderer: "gl",
  doubleBuffer: false,
  glOptions: void 0,
  forceRenderOnZooming: true
};
class Layer extends maptalks__namespace.TileLayer {
  constructor(id, source, opts) {
    super(id, {
      ...opts,
      urlTemplate: "custom",
      tileSize: source.tileSize,
      repeatWorld: source.wrapX ? "x" : false
    });
    this._needsUpdate = true;
    this._coordCache = {};
    this.tempLayer = new maptalks__namespace.TileLayer(id + "_temp", {
      ...opts,
      minZoom: source.minZoom,
      maxZoom: source.maxZoom,
      urlTemplate: "custom",
      tileSize: source.tileSize,
      repeatWorld: source.wrapX ? "x" : false
    });
    this.tempLayer.on("renderercreate", (e) => {
      e.renderer.loadTile = function loadTile(tile) {
        return tile;
      };
      e.renderer.deleteTile = (tile) => {
        if (!tile || !tile.image) {
          return;
        }
        tile.image.onload = null;
        tile.image.onerror = null;
      };
      e.renderer.loadTileImage = (img, url, key) => {
      };
    });
    this.source = source;
    this.type = "VeLayer";
  }
  onMoveStart() {
    if (this.layer) {
      this.layer.update();
      this.layer.moveStart();
    }
  }
  onMoving() {
    this.layer?.update();
  }
  onMoveEnd() {
    if (this.layer) {
      this.layer.update();
      this.layer.moveEnd();
    }
  }
  onDragRotateStart() {
    if (this.layer) {
      this.layer.moveStart();
    }
  }
  onDragRotateEnd() {
    if (this.layer) {
      this.layer.moveEnd();
    }
  }
  onResize() {
    this.update();
  }
  onZoomStart() {
    this.layer?.update();
  }
  onZooming() {
    if (this.layer) {
      this.layer.update();
      this.layer.handleZoom();
    }
  }
  onZoomEnd() {
    this.layer?.update();
  }
  update() {
    if (this.layer) {
      this.layer.update();
    }
  }
  getWorlds() {
    const map = this.getMap();
    const projObject = map.getProjection().fullExtent;
    const projectionExtent = [projObject.left, projObject.bottom, projObject.right, projObject.top];
    const projExtent = map.getProjExtent();
    const extent = [projExtent.xmin, projExtent.ymin, projExtent.xmax, projExtent.ymax];
    const worldWidth = projectionExtent[2] - projectionExtent[0];
    const res = getGLRes(map);
    const p1 = map._prjToPointAtRes(new maptalks__namespace.Coordinate([projectionExtent[0], projectionExtent[3]]), res);
    const p2 = map._prjToPointAtRes(new maptalks__namespace.Coordinate([projectionExtent[2], projectionExtent[1]]), res);
    const projWorldWidth = Math.abs(p1.x - p2.x);
    let startX = projectionExtent[0];
    let world = 0;
    const result = [
      {
        world: 0,
        offset: 0,
        xmin: p1.x,
        xmax: p2.x
      }
    ];
    while (startX > extent[0]) {
      --world;
      result.push({
        world,
        offset: world * projWorldWidth,
        xmin: p1.x + world * projWorldWidth,
        xmax: p2.x + world * projWorldWidth
      });
      startX -= worldWidth;
    }
    world = 0;
    startX = projectionExtent[2];
    while (startX < extent[2]) {
      ++world;
      result.push({
        world,
        offset: world * projWorldWidth,
        xmin: p1.x + world * projWorldWidth,
        xmax: p2.x + world * projWorldWidth
      });
      startX += worldWidth;
    }
    this.projWorldWidth = projWorldWidth;
    return result.sort((a, b) => a.world - b.world);
  }
  calcWitchWorld(center, worlds) {
    for (let i = 0; i < worlds.length; i++) {
      if (inRange(center[0], worlds[i].xmin, worlds[i].xmax)) {
        return worlds[i].world;
      }
    }
    return worlds.find((w) => w.world === 0)?.world;
  }
  getClipMask() {
    return this.options.mask;
  }
  processMask() {
    const map = this.getMap();
    if (map && this.options.mask) {
      const r = getGLRes(map);
      const mask = this.options.mask;
      const data = mask.data;
      rewind(data, true);
      const tr = (coords) => {
        const mercatorCoordinates = [];
        for (let i2 = 0; i2 < coords.length; i2++) {
          const coord = coords[i2];
          const p = coordinateToPoint(map, new maptalks__namespace.Coordinate(coord), r);
          mercatorCoordinates.push([p.x, p.y]);
        }
        return mercatorCoordinates;
      };
      const features = data.features;
      const len = features.length;
      let i = 0;
      const fs = [];
      for (; i < len; i++) {
        const feature = features[i];
        const coordinates = feature.geometry.coordinates;
        const type = feature.geometry.type;
        if (type === "Polygon") {
          fs.push({
            type: "Feature",
            properties: {},
            geometry: {
              type: "Polygon",
              coordinates: feature.geometry.coordinates.map((c) => tr(c))
            }
          });
        } else if (type === "MultiPolygon") {
          const css = [];
          for (let k = 0; k < coordinates.length; k++) {
            const coordinate = coordinates[k];
            const cs = [];
            for (let n = 0; n < coordinate.length; n++) {
              cs.push(tr(coordinates[k][n]));
            }
            css.push(cs);
          }
          fs.push({
            type: "Feature",
            properties: {},
            geometry: {
              type: "MultiPolygon",
              coordinates: css
            }
          });
        }
      }
      return {
        data: windGlCore.polygon2buffer(fs),
        type: mask.type
      };
    }
  }
  setClipMask(mask) {
    this.options.mask = Object.assign({}, this.options.mask, mask);
    if (this.layer) {
      this.layer.setMask(this.processMask());
    }
  }
  // eslint-disable-next-line
  prepareToDraw(gl, scene) {
    const opt = this.options;
    const map = this.getMap();
    const renderer = this._getRenderer();
    if (!map || !renderer) {
      return false;
    }
    this.getWorlds();
    this.layer = new windGlCore.BaseLayer(
      this.source,
      {
        renderer: renderer.context,
        scene: renderer.scene
      },
      {
        renderType: opt.renderType,
        renderFrom: opt.renderFrom,
        styleSpec: opt.styleSpec,
        displayRange: opt.displayRange,
        widthSegments: opt.widthSegments,
        heightSegments: opt.heightSegments,
        wireframe: opt.wireframe,
        picking: opt.picking,
        mask: this.processMask(),
        getZoom: () => map.getZoom(),
        triggerRepaint: () => {
          renderer.setToRedraw();
        },
        flipY: true,
        glScale: () => this.projWorldWidth,
        zoomScale: () => 2,
        // 由于核心库中计算瓦片坐标是从上到下是递增的，但是 maptalk 是从上到下是递减的
        getTileProjSize: (z, tiles) => {
          const t = tiles.find((tile) => tile.z === z);
          if (t) {
            return [t.projTileBounds.right - t.projTileBounds.left, t.projTileBounds.top - t.projTileBounds.bottom];
          }
          return [this.projWorldWidth, this.projWorldWidth];
        },
        getViewTiles: (source) => {
          let { type } = source;
          const worlds = this.getWorlds();
          type = type !== windGlCore.LayerSourceType.timeline ? type : source.privateType;
          const wrapTiles = [];
          const r = getGLRes(map);
          if (type === windGlCore.LayerSourceType.image) {
            const { coordinates } = source;
            const [min, max] = [
              [coordinates[0][0], coordinates[2][1]],
              [coordinates[1][0], coordinates[0][1]]
            ];
            const p1 = coordinateToPoint(map, new maptalks__namespace.Coordinate([min[0], max[1]]), r);
            const p2 = coordinateToPoint(map, new maptalks__namespace.Coordinate([max[0], min[1]]), r);
            const x = 0;
            const y = 0;
            const z = 0;
            if (this.source.wrapX) {
              for (let i = 0; i < worlds.length; i++) {
                const { world, offset } = worlds[i];
                wrapTiles.push(
                  new windGlCore.TileID(z, world, z, x, y, {
                    getTileBounds: () => [min[0], min[1], max[0], max[1]],
                    getTileProjBounds: () => ({
                      left: p1.x + offset,
                      top: p1.y,
                      right: p2.x + offset,
                      bottom: p2.y
                    })
                  })
                );
              }
            } else {
              const wrap = 0;
              wrapTiles.push(
                new windGlCore.TileID(z, wrap, z, x, y, {
                  getTileBounds: () => [min[0], min[1], max[0], max[1]],
                  getTileProjBounds: () => ({
                    left: p1.x,
                    top: p1.y,
                    right: p2.x,
                    bottom: p2.y
                  })
                })
              );
            }
          } else if (type === windGlCore.LayerSourceType.tile) {
            const { tileGrids } = this.tempLayer.getTiles();
            const { tiles } = tileGrids[0];
            for (let i = 0; i < tiles.length; i++) {
              const tile = tiles[i];
              const res = map._getResolution(tile.z);
              const tileConfig = this._getTileConfig();
              const tileExtent = tileConfig.getTilePrjExtent(tile.x, tile.y, res);
              let max = tileExtent.getMax();
              let min = tileExtent.getMin();
              const projection = map.getProjection();
              min = projection.unproject(min);
              max = projection.unproject(max);
              const { extent2d, offset } = tile;
              const scale = tile._glScale = tile._glScale || tile.res / map.getGLRes();
              const point1 = TILE_POINT.set(extent2d.xmin - offset[0], extent2d.ymax - offset[1]);
              const x1 = point1.x * scale;
              const y1 = point1.y * scale;
              const point2 = TILE_POINT.set(extent2d.xmax - offset[0], extent2d.ymin - offset[1]);
              const x2 = point2.x * scale;
              const y2 = point2.y * scale;
              const wrap = this.calcWitchWorld([(x2 - x1) / 2 + x1, (y2 - y1) / 2 + y1], worlds);
              wrapTiles.push(
                new windGlCore.TileID(tile.z, wrap, tile.z, tile.x, tile.y, {
                  getTileBounds: () => [min.x, min.y, max.x, max.y],
                  getTileProjBounds: () => ({
                    left: x1,
                    top: y1,
                    right: x2,
                    bottom: y2
                  })
                })
              );
            }
          }
          return wrapTiles;
        },
        getExtent: () => {
          const projExtent = map.getProjExtent()._scale(1 / getGLRes(map));
          const r = getGLRes(map);
          const p1 = coordinateToPoint(map, new maptalks__namespace.Coordinate([0, -90]), r);
          const p2 = coordinateToPoint(map, new maptalks__namespace.Coordinate([0, 90]), r);
          return [
            projExtent.xmin,
            projExtent.ymin < p1.y ? p1.y : projExtent.ymin,
            projExtent.xmax,
            projExtent.ymax > p2.y ? p2.y : projExtent.ymax
          ];
        },
        getGridTiles: (_) => {
          const { tileGrids } = this.getTiles();
          const { tiles } = tileGrids[0];
          const wrapTiles = [];
          const worlds = this.getWorlds();
          for (let i = 0; i < tiles.length; i++) {
            const tile = tiles[i];
            const res = map._getResolution(tile.z);
            const tileConfig = this._getTileConfig();
            const tileExtent = tileConfig.getTilePrjExtent(tile.x, tile.y, res);
            let max = tileExtent.getMax();
            let min = tileExtent.getMin();
            const projection = map.getProjection();
            min = projection.unproject(min);
            max = projection.unproject(max);
            const { extent2d, offset } = tile;
            const scale = tile._glScale = tile._glScale || tile.res / map.getGLRes();
            const point1 = TILE_POINT.set(extent2d.xmin - offset[0], extent2d.ymax - offset[1]);
            const x1 = point1.x * scale;
            const y1 = point1.y * scale;
            const point2 = TILE_POINT.set(extent2d.xmax - offset[0], extent2d.ymin - offset[1]);
            const x2 = point2.x * scale;
            const y2 = point2.y * scale;
            const wrap = this.calcWitchWorld([(x2 - x1) / 2 + x1, (y2 - y1) / 2 + y1], worlds);
            wrapTiles.push(
              new windGlCore.TileID(tile.z, wrap, tile.z, tile.x, tile.y, {
                getTileBounds: () => [min.x, min.y, max.x, max.y],
                getTileProjBounds: () => ({
                  left: x1,
                  top: y1,
                  right: x2,
                  bottom: y2
                })
              })
            );
          }
          return wrapTiles;
        }
      }
    );
    map.on("viewchange", this.update, this);
    this.update();
  }
  isRendering() {
    const map = this.getMap();
    if (!map) {
      return false;
    }
    return map.isInteracting() || map.isAnimating();
  }
  draw(gl, view, scene, camera, timeStamp, context) {
    this.renderScene(context, this);
  }
  drawOnInteracting(gl, view, scene, camera, event, timeStamp, context) {
    this.renderScene(context, this);
  }
  getObjects() {
    const scene = this.getScene();
    if (!scene) {
      return [];
    }
    const meshes = [];
    for (let i = 0, len = scene.children.length; i < len; i++) {
      const child = scene.children[i];
      if (child) {
        meshes.push(child);
      }
    }
    return meshes;
  }
  clear() {
    return this.clearObject();
  }
  clearObject() {
    const scene = this.getScene();
    if (!scene) {
      return this;
    }
    for (let i = scene.children.length - 1; i >= 0; i--) {
      const child = scene.children[i];
      if (child) {
        scene.remove(child);
      }
    }
    return this;
  }
  getCamera() {
    const renderer = this._getRenderer();
    if (renderer) {
      return renderer.camera;
    }
    return null;
  }
  getScene() {
    const renderer = this._getRenderer();
    if (renderer) {
      return renderer.scene;
    }
    return null;
  }
  renderScene(context, layer) {
    const renderer = this._getRenderer();
    if (renderer) {
      renderer.clearCanvas();
      renderer.renderScene(context);
      if (!layer) {
        renderer.setToRedraw();
      }
    }
    return this;
  }
  onAdd() {
    super.onAdd();
    const map = this.map || this.getMap();
    if (!map)
      return this;
    map.addLayer(this.tempLayer);
    this._needsUpdate = true;
    return this;
  }
  onRemove() {
    super.onRemove();
    const map = this.map || this.getMap();
    if (!map)
      return this;
    map?.off("viewchange", this.update, this);
    this.clear();
    if (this.layer) {
      this.layer = null;
    }
    return this;
  }
  getEvents() {
    return {
      _zoomstart: this.onZoomStart,
      _zooming: this.onZooming,
      _zoomend: this.onZoomEnd,
      _resize: this.onResize,
      _movestart: this.onMoveStart,
      _moving: this.onMoving,
      _moveend: this.onMoveEnd,
      _dragrotatestart: this.onDragRotateStart,
      // '_dragrotating': this.onDragRotating,
      _dragrotateend: this.onDragRotateEnd
      // _spatialreferencechange: this.onSpatialReferenceChange,
    };
  }
}
Layer.mergeOptions(options);
Layer.registerRenderer("gl", VeRenderer);

Object.defineProperty(exports, "DecodeType", {
  enumerable: true,
  get: function () { return windGlCore.DecodeType; }
});
Object.defineProperty(exports, "ImageSource", {
  enumerable: true,
  get: function () { return windGlCore.ImageSource; }
});
Object.defineProperty(exports, "LayerSourceType", {
  enumerable: true,
  get: function () { return windGlCore.LayerSourceType; }
});
Object.defineProperty(exports, "MaskType", {
  enumerable: true,
  get: function () { return windGlCore.MaskType; }
});
Object.defineProperty(exports, "RenderFrom", {
  enumerable: true,
  get: function () { return windGlCore.RenderFrom; }
});
Object.defineProperty(exports, "RenderType", {
  enumerable: true,
  get: function () { return windGlCore.RenderType; }
});
Object.defineProperty(exports, "TileID", {
  enumerable: true,
  get: function () { return windGlCore.TileID; }
});
Object.defineProperty(exports, "TileSource", {
  enumerable: true,
  get: function () { return windGlCore.TileSource; }
});
Object.defineProperty(exports, "TimelineSource", {
  enumerable: true,
  get: function () { return windGlCore.TimelineSource; }
});
Object.defineProperty(exports, "configDeps", {
  enumerable: true,
  get: function () { return windGlCore.configDeps; }
});
Object.defineProperty(exports, "Field", {
  enumerable: true,
  get: function () { return windCore.Field; }
});
Object.defineProperty(exports, "WindCore", {
  enumerable: true,
  get: function () { return windCore.WindCore; }
});
Object.defineProperty(exports, "formatData", {
  enumerable: true,
  get: function () { return windCore.formatData; }
});
exports.Layer = Layer;
exports.WindLayer = MaptalksWind;
//# sourceMappingURL=maptalks-wind.cjs.js.map
